import torch
import math
from typing import Dict, List

def get_surprise_from_grads(gradients: List[torch.Tensor]) -> float:
    with torch.no_grad():
        total_norm_sq = sum(p.norm(2).item() ** 2 for p in gradients if p is not None)
    return math.sqrt(total_norm_sq)

def get_entropy_from_logits(logits: torch.Tensor) -> float:
    probs = torch.softmax(logits, dim=-1)
    entropy = -torch.sum(probs * torch.log(probs + 1e-9), dim=-1)
    return entropy.mean().item()

def calculate_pi(
    epsilon: float,
    tau: float,
    surprise: float,
    alpha: float = 1.0,
    gamma: float = 0.5
) -> Dict[str, float]:
    normalized_error = epsilon / (tau + 1e-9)
    cognitive_cost = (1 - gamma) * normalized_error + gamma * surprise
    pi_score = math.exp(-alpha * cognitive_cost)

    return {
        "pi_score": pi_score,
        "normalized_error": normalized_error,
        "cognitive_cost": cognitive_cost,
        "epsilon": epsilon,
        "tau": tau,
        "surprise": surprise,
    }

class SigmaPI:
    def __init__(self, alpha: float = 1.0, gamma: float = 0.5):
        self.alpha = alpha
        self.gamma = gamma

    def calculate(
        self,
        model: torch.nn.Module,
        loss_epsilon: torch.Tensor,
        logits: torch.Tensor
    ) -> Dict[str, float]:
        model_grads = [p.grad for p in model.parameters() if p.grad is not None]
        if not model_grads:
            raise ValueError("No gradients found. Please call loss.backward() before calculate().")

        epsilon = loss_epsilon.item()
        model_grads = [p.grad for p in model.parameters() if p.grad is not None]

        tau = get_entropy_from_logits(logits)
        surprise = get_surprise_from_grads(model_grads)

        pi_metrics = calculate_pi(
            epsilon=epsilon,
            tau=tau,
            surprise=surprise,
            alpha=self.alpha,
            gamma=self.gamma
        )
        return pi_metrics



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import os
import argparse
import importlib.util
from main import SigmaPI
from test.utils.training import train, validate
from test.utils.plotting import plot_metrics
from test.models.vit import VisionTransformer, MoEVisionTransformer
from torchvision import datasets, transforms
from datetime import datetime
from typing import Dict, Any

class Tee(object):
    def __init__(self, *files):
        self.files = files
    def write(self, obj):
        for f in self.files:
            f.write(obj)
            f.flush()
    def flush(self):
        for f in self.files:
            f.flush()
    def isatty(self):
        return False

def setup_experiment(config: Dict[str, Any], model_name_suffix: str = "", checkpoint_path: str = None):
    model_cfg = config.model_config
    train_cfg = config.train_config
    pi_cfg = config.pi_config

    device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    if device.type == 'cpu':
        raise RuntimeError("CUDA not available, exiting.")

    model_type = model_cfg.get('model_type') # Use .get() to avoid popping
    if model_type == 'base':
        model = VisionTransformer(**model_cfg).to(device)
    elif model_type in ['moe', 'gbp_moe']:
        model = MoEVisionTransformer(**model_cfg).to(device)
    else:
        raise ValueError(f"Unknown model_type: {model_type}")

    if checkpoint_path:
        print(f"Loading checkpoint from: {checkpoint_path}")
        model.load_state_dict(torch.load(checkpoint_path))

    total_params = sum(p.numel() for p in model.parameters() if p.requires_grad)
    
    # Construct a clean model name based on the config file name
    config_name = model_name_suffix.split('/')[-1].replace('.py', '')
    model_name = f"{config_name}{model_name_suffix}"
    print(f"Model: {model_name}")
    print(f"Total Trainable Parameters: {total_params/1e6:.2f}M")

    optimizer = optim.AdamW(model.parameters(), lr=train_cfg['learning_rate'], weight_decay=train_cfg['weight_decay'])
    loss_fn = nn.CrossEntropyLoss()
    pi_monitor = SigmaPI(**pi_cfg)

    return model, optimizer, loss_fn, pi_monitor, device, model_name

def get_dataloaders(dataset_name: str, batch_size: int, num_workers: int = 8):
    transform_train = transforms.Compose([
        transforms.RandomCrop(32, padding=4),
        transforms.RandomHorizontalFlip(),
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])
    transform_test = transforms.Compose([
        transforms.ToTensor(),
        transforms.Normalize((0.4914, 0.4822, 0.4465), (0.2023, 0.1994, 0.2010)),
    ])

    if dataset_name.upper() == 'CIFAR10':
        data_dir = "temp_data/CIFAR10"
        os.makedirs(data_dir, exist_ok=True)
        is_downloaded = os.path.exists(os.path.join(data_dir, 'cifar-10-batches-py'))
        train_dataset = datasets.CIFAR10(data_dir, train=True, download=not is_downloaded, transform=transform_train)
        val_dataset = datasets.CIFAR10(data_dir, train=False, download=not is_downloaded, transform=transform_test)
    elif dataset_name.upper() == 'SVHN':
        data_dir = "temp_data/SVHN"
        os.makedirs(data_dir, exist_ok=True)
        train_file = os.path.join(data_dir, 'train_32x32.mat')
        test_file = os.path.join(data_dir, 'test_32x32.mat')
        train_dataset = datasets.SVHN(data_dir, split='train', download=not os.path.exists(train_file), transform=transform_train)
        val_dataset = datasets.SVHN(data_dir, split='test', download=not os.path.exists(test_file), transform=transform_test)
    else:
        raise ValueError(f"Unknown dataset: {dataset_name}")

    train_loader = DataLoader(train_dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, pin_memory=True)
    val_loader = DataLoader(val_dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers, pin_memory=True)
    
    return train_loader, val_loader

def execute_training_loop(
    model, optimizer, loss_fn, pi_monitor, device, model_name,
    train_loader, val_loaders: Dict[str, DataLoader],
    epochs: int, accumulation_steps: int, output_dir: str,
    use_gbp: bool,
    start_epoch: int = 1, global_step: int = 0,
    train_fn_kwargs: Dict[str, Any] = {}
):
    metric_keys = ['loss', 'acc', 'pi', 'surprise', 'tau']
    step_metrics = {f'train_{key}': [] for key in metric_keys}
    epoch_metrics = {f'train_{key}': [] for key in metric_keys}
    for val_set_name in val_loaders.keys():
        for key in metric_keys:
            epoch_metrics[f'{val_set_name}_{key}'] = []

    for epoch in range(start_epoch, start_epoch + epochs):
        print(f"\n--- Epoch {epoch} ---")
        
        train_output = train(
            model, device, train_loader, optimizer, epoch, loss_fn, 
            pi_monitor, step_metrics, epoch_metrics, global_step, 
            accumulation_steps, use_gbp=use_gbp, **train_fn_kwargs
        )

        if use_gbp:
            global_step, gbp_surprise_values, gbp_decisions = train_output
            if 'gbp_surprise_values' not in epoch_metrics:
                epoch_metrics['gbp_surprise_values'] = []
                epoch_metrics['gbp_decisions'] = []
            epoch_metrics['gbp_surprise_values'].extend(gbp_surprise_values)
            epoch_metrics['gbp_decisions'].extend(gbp_decisions)
        else:
            global_step = train_output

        for name, loader in val_loaders.items():
            val_loss, val_acc, val_pi, val_surprise, val_tau = validate(model, device, loader, loss_fn, pi_monitor, dataset_name=name)
            epoch_metrics[f'{name}_loss'].append((global_step - 1, val_loss))
            epoch_metrics[f'{name}_acc'].append((global_step - 1, val_acc))
            epoch_metrics[f'{name}_pi'].append((global_step - 1, val_pi))
            epoch_metrics[f'{name}_surprise'].append((global_step - 1, val_surprise))
            epoch_metrics[f'{name}_tau'].append((global_step - 1, val_tau))

    checkpoint_dir = os.path.join(output_dir, 'checkpoints')
    os.makedirs(checkpoint_dir, exist_ok=True)
    checkpoint_name = f"{model_name}-epoch_{start_epoch + epochs - 1}.pth"
    checkpoint_path = os.path.join(checkpoint_dir, checkpoint_name)
    torch.save(model.state_dict(), checkpoint_path)
    print(f"\nCheckpoint saved to: {checkpoint_path}")

    plot_metrics_kwargs = {}
    if 'gbp_surprise_values' in epoch_metrics:
        plot_metrics_kwargs['gbp_surprise_values'] = epoch_metrics['gbp_surprise_values']
        plot_metrics_kwargs['gbp_decisions'] = epoch_metrics['gbp_decisions']

    plot_metrics(step_metrics, epoch_metrics, output_dir, model_name=model_name, **plot_metrics_kwargs)
    print(f"\nPlots saved to: {os.path.abspath(output_dir)}")
    
    return model, global_step

def run_experiment(config, config_path: str, task_name_suffix: str = "", checkpoint_path: str = None):
    train_cfg = config.train_config
    log_dir = os.path.join(train_cfg['output_dir'], 'log')
    os.makedirs(log_dir, exist_ok=True)
    current_time = datetime.now().strftime("%Y%m%d-%H%M%S")
    config_name = os.path.basename(config_path).replace('.py', '')
    log_file_name = f"{current_time}-{config_name}{task_name_suffix}.log"
    log_file_path = os.path.join(log_dir, log_file_name)

    original_stdout = sys.stdout
    log_file = open(log_file_path, 'w')
    sys.stdout = Tee(original_stdout, log_file)
    print(f"Logging output to: {log_file_path}")

    try:
        model, optimizer, loss_fn, pi_monitor, device, model_name = setup_experiment(config, model_name_suffix=f"{task_name_suffix}", checkpoint_path=checkpoint_path)
        
        train_loader, val_loader = get_dataloaders('CIFAR10', train_cfg['batch_size'])
        _, ood_val_loader = get_dataloaders('SVHN', train_cfg['batch_size'])

        val_loaders = {
            "CIFAR10_Val": val_loader,
            "SVHN_OOD_Val": ood_val_loader
        }

        execute_training_loop(
            model=model, optimizer=optimizer, loss_fn=loss_fn, pi_monitor=pi_monitor,
            device=device, model_name=model_name,
            train_loader=train_loader, val_loaders=val_loaders,
            epochs=train_cfg['epochs'], accumulation_steps=train_cfg['accumulation_steps'],
            use_gbp=(config.model_config.get('model_type') == 'gbp_moe'),
            output_dir=train_cfg['output_dir'],
            train_fn_kwargs=train_cfg.get('train_fn_kwargs', {})
        )
    finally:
        sys.stdout = original_stdout
        log_file.close()
        print(f"Training completed. Log saved to: {log_file_path}")

def main():
    parser = argparse.ArgumentParser(description="Run a ViT experiment from a configuration file.")
    parser.add_argument('--config', type=str, required=True, help="Path to the configuration file.")
    parser.add_argument('--task', type=str, default="single", help="Task name for logging (e.g., 'CIFAR10', 'continual').")
    parser.add_argument('--checkpoint_path', type=str, default=None, help="Path to a model checkpoint to load.")
    args = parser.parse_args()

    # Load config module
    spec = importlib.util.spec_from_file_location(name="config", location=args.config)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)
    
    run_experiment(config, config_path=args.config, task_name_suffix=f"-{args.task}", checkpoint_path=args.checkpoint_path)

if __name__ == "__main__":
    main()



import torch
import torch.nn as nn
import torch.optim as optim
from torchvision import datasets, transforms
from torch.utils.data import DataLoader
import os
from main import SigmaPI
from test.utils.plotting import plot_metrics

def train(model, device, train_loader, optimizer, epoch, loss_fn, pi_monitor, 
          step_metrics, epoch_metrics, global_step, accumulation_steps, 
          use_gbp=False, initial_surprise_ema=None):
    model.train()
    optimizer.zero_grad()

    # GBP specific state
    ema_surprise = initial_surprise_ema
    ema_surprise_sq = initial_surprise_ema ** 2 if initial_surprise_ema is not None else None
    alpha = 0.1

    epoch_summary = {
        'loss': [], 'acc': [], 'pi': [], 'surprise': [], 'tau': [],
        'consolidate': [], 'ignore': [], 'reject': [],
        'surprise_values': [], 'decisions': []
    }

    for batch_idx, (data, target) in enumerate(train_loader):
        data, target = data.to(device), target.to(device)

        output = model(data)
        # Handle both standard and MoE model outputs
        logits = output[0] if isinstance(output, tuple) else output
        all_gating_logits = output[1] if isinstance(output, tuple) else None

        loss = loss_fn(logits, target)
        loss_normalized = loss / accumulation_steps
        loss_normalized.backward()

        pred = logits.argmax(dim=1, keepdim=True)
        correct = pred.eq(target.view_as(pred)).sum().item()
        accuracy = 100. * correct / len(data)

        if (batch_idx + 1) % accumulation_steps == 0:
            if use_gbp:
                if all_gating_logits is None:
                    raise ValueError("GBP requires a model that returns gating logits.")
                model.zero_inactive_expert_grads(all_gating_logits)
            
            pi_metrics = pi_monitor.calculate(model, loss, logits)
            
            if use_gbp:
                surprise = torch.tensor(pi_metrics['surprise'], device=device)
                if ema_surprise is None:
                    ema_surprise = surprise
                    ema_surprise_sq = surprise ** 2
                else:
                    ema_surprise = alpha * surprise + (1 - alpha) * ema_surprise
                    ema_surprise_sq = alpha * (surprise ** 2) + (1 - alpha) * ema_surprise_sq
                
                variance = ema_surprise_sq - ema_surprise ** 2
                std_dev = torch.sqrt(torch.clamp(variance, min=0))
                
                s_low = ema_surprise - 2.5 * std_dev
                s_high = ema_surprise + 2.5 * std_dev

                decision = "IGNORE"
                if s_low <= surprise < s_high:
                    decision = "CONSOLIDATE"
                    optimizer.step()
                elif surprise >= s_high:
                    decision = "REJECT"
                
                epoch_summary[decision.lower()].append(1)
                epoch_summary['surprise_values'].append((global_step, surprise.item()))
                epoch_summary['decisions'].append((global_step, decision))
            else:
                torch.nn.utils.clip_grad_norm_(model.parameters(), 1.0)
                optimizer.step()

            optimizer.zero_grad()

            # Log metrics for both modes
            step_metrics['train_loss'].append((global_step, loss.item()))
            step_metrics['train_acc'].append((global_step, accuracy))
            step_metrics['train_pi'].append((global_step, pi_metrics['pi_score']))
            step_metrics['train_surprise'].append((global_step, pi_metrics['surprise']))
            step_metrics['train_tau'].append((global_step, pi_metrics['tau']))
            
            # Log metrics for epoch summary
            epoch_summary['loss'].append(loss.item())
            epoch_summary['acc'].append(accuracy)
            epoch_summary['pi'].append(pi_metrics['pi_score'])
            epoch_summary['surprise'].append(pi_metrics['surprise'])
            epoch_summary['tau'].append(pi_metrics['tau'])

            global_step += 1
        
    # Averaging and logging epoch summaries
    avg_metrics = {key: sum(vals) / len(vals) if vals else 0 for key, vals in epoch_summary.items() if key not in ['consolidate', 'ignore', 'reject', 'surprise_values', 'decisions']}
    
    for key, avg_val in avg_metrics.items():
        epoch_metrics[f'train_{key}'].append((global_step - 1, avg_val))

    print(f"Train Epoch {epoch} Summary: Avg loss: {avg_metrics['loss']:.4f}, Avg Accuracy: {avg_metrics['acc']:.2f}%, Avg PI: {avg_metrics['pi']:.4f}, Avg Surprise: {avg_metrics['surprise']:.4f}, Avg Tau: {avg_metrics['tau']:.4f}")
    
    if use_gbp:
        consolidate_count = len(epoch_summary['consolidate'])
        ignore_count = len(epoch_summary['ignore'])
        reject_count = len(epoch_summary['reject'])
        total_decisions = consolidate_count + ignore_count + reject_count
        if total_decisions > 0:
            print(f"Decision Stats: Consolidate: {100*consolidate_count/total_decisions:.1f}%, Ignore: {100*ignore_count/total_decisions:.1f}%, Reject: {100*reject_count/total_decisions:.1f}%")
        return global_step, epoch_summary['surprise_values'], epoch_summary['decisions']
    
    return global_step

def validate(model, device, val_loader, loss_fn, pi_monitor, dataset_name="Validation"):
    model.eval()
    total_loss, correct = 0, 0
    all_pi_scores, all_surprises, all_taus = [], [], []

    for data, target in val_loader:
        data, target = data.to(device), target.to(device)
        
        # Enable gradients for PI calculation
        with torch.enable_grad():
            output = model(data)
            logits = output[0] if isinstance(output, tuple) else output
            loss_epsilon = loss_fn(logits, target)
            
            model.zero_grad()
            loss_epsilon.backward()
            
            pi_metrics = pi_monitor.calculate(model, loss_epsilon, logits)
            all_pi_scores.append(pi_metrics['pi_score'])
            all_surprises.append(pi_metrics['surprise'])
            all_taus.append(pi_metrics['tau'])

        # No-grad context for accuracy and loss accumulation
        with torch.no_grad():
            total_loss += loss_epsilon.item()
            pred = logits.argmax(dim=1, keepdim=True)
            correct += pred.eq(target.view_as(pred)).sum().item()

    avg_loss = total_loss / len(val_loader)
    accuracy = 100. * correct / len(val_loader.dataset)
    avg_pi = sum(all_pi_scores) / len(all_pi_scores) if all_pi_scores else 0
    avg_surprise = sum(all_surprises) / len(all_surprises) if all_surprises else 0
    avg_tau = sum(all_taus) / len(all_taus) if all_taus else 0

    print(f"{dataset_name} set: Avg loss: {avg_loss:.4f}, Accuracy: {accuracy:.2f}%, Avg PI: {avg_pi:.4f}, Avg Surprise: {avg_surprise:.4f}, Avg Tau: {avg_tau:.4f}")
    return avg_loss, accuracy, avg_pi, avg_surprise, avg_tau



import os
import sys
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '..')))

import argparse
import importlib.util
import torch
from datetime import datetime

from test.run_experiment import setup_experiment, get_dataloaders, execute_training_loop, Tee
from test.utils.training import validate

def run_continual_experiment(config_path: str, checkpoint_path: str = None, start_stage: int = 1):
    spec = importlib.util.spec_from_file_location(name="config", location=config_path)
    config = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(config)

    train_cfg = config.train_config
    log_dir = os.path.join(train_cfg['output_dir'], 'log')
    os.makedirs(log_dir, exist_ok=True)
    current_time = datetime.now().strftime("%Y%m%d-%H%M%S")
    config_name = os.path.basename(config_path).replace('.py', '')
    log_file_name = f"{current_time}-{config_name}-continual.log"
    log_file_path = os.path.join(log_dir, log_file_name)

    original_stdout = sys.stdout
    log_file = open(log_file_path, 'w')
    sys.stdout = Tee(original_stdout, log_file)
    print(f"Logging output to: {log_file_path}")

    try:
        task_suffix = "-continual"
        model, optimizer, loss_fn, pi_monitor, device, model_name = setup_experiment(
            config, model_name_suffix=task_suffix, checkpoint_path=checkpoint_path
        )
        
        global_step = 0
        
        # --- STAGE 1: CIFAR-10 ---
        if start_stage <= 1:
            print("\n" + "="*50)
            print(" " * 15 + "STAGE 1: Training on CIFAR-10")
            print("="*50 + "\n")
            
            cifar_train_loader, cifar_val_loader = get_dataloaders('CIFAR10', train_cfg['batch_size'])
            stage1_val_loaders = {"CIFAR10_Val": cifar_val_loader}
            
            model, global_step = execute_training_loop(
                model=model, optimizer=optimizer, loss_fn=loss_fn, pi_monitor=pi_monitor,
                device=device, model_name=model_name,
                train_loader=cifar_train_loader, val_loaders=stage1_val_loaders,
                epochs=20,
                use_gbp=(config.model_config.get('model_type') == 'gbp_moe'),
                accumulation_steps=train_cfg['accumulation_steps'],
                output_dir=train_cfg['output_dir'],
                train_fn_kwargs=train_cfg.get('train_fn_kwargs', {})
            )
        else:
            print("\nSkipping Stage 1: Training on CIFAR-10 as requested.")

        # --- STAGE 2: SVHN ---
        if start_stage <= 2:
            print("\n" + "="*50)
            print(" " * 15 + "STAGE 2: Training on SVHN")
            print("="*50 + "\n")

            svhn_train_loader, svhn_val_loader = get_dataloaders('SVHN', train_cfg['batch_size'])
            # We need cifar_val_loader for the final comparison
            _, cifar_val_loader = get_dataloaders('CIFAR10', train_cfg['batch_size'])
            
            stage2_val_loaders = {
                "CIFAR10_Val": cifar_val_loader,
                "SVHN_Val": svhn_val_loader
            }

            model, _ = execute_training_loop(
                model=model, optimizer=optimizer, loss_fn=loss_fn, pi_monitor=pi_monitor,
                device=device, model_name=model_name,
                train_loader=svhn_train_loader, val_loaders=stage2_val_loaders,
                epochs=10,
                use_gbp=(config.model_config.get('model_type') == 'gbp_moe'),
                accumulation_steps=train_cfg['accumulation_steps'],
                output_dir=train_cfg['output_dir'],
                start_epoch=21,
                global_step=global_step,
                train_fn_kwargs=train_cfg.get('train_fn_kwargs', {})
            )

        # --- FINAL EVALUATION ---
        print("\n" + "="*50)
        print(" " * 18 + "FINAL EVALUATION")
        print("="*50 + "\n")

        print(f"--- Final performance for model: {model_name} ---")
        
        _, cifar_val_loader = get_dataloaders('CIFAR10', train_cfg['batch_size'])
        _, svhn_val_loader = get_dataloaders('SVHN', train_cfg['batch_size'])
        
        validate(model, device, cifar_val_loader, loss_fn, pi_monitor, dataset_name="Final CIFAR-10 Val")
        validate(model, device, svhn_val_loader, loss_fn, pi_monitor, dataset_name="Final SVHN Val")

    finally:
        sys.stdout = original_stdout
        log_file.close()
        print(f"Training completed. Log saved to: {log_file_path}")

def main():
    parser = argparse.ArgumentParser(description="Run a continual learning experiment (CIFAR-10 -> SVHN).")
    parser.add_argument('--config', type=str, required=True, help="Path to the model configuration file.")
    parser.add_argument('--checkpoint_path', type=str, default=None, help="Path to a model checkpoint to load and continue training.")
    parser.add_argument('--start_stage', type=int, default=1, choices=[1, 2], help="Start from a specific stage (1: CIFAR, 2: SVHN).")
    args = parser.parse_args()
    run_continual_experiment(args.config, args.checkpoint_path, args.start_stage)

if __name__ == "__main__":
    main()



import matplotlib.pyplot as plt
import os
from datetime import datetime
from typing import Dict, List, Tuple, Optional

def plot_gbp_decisions(
    gbp_surprise_values: List[Tuple[int, float]],
    gbp_decisions: List[Tuple[int, str]],
    output_dir: str,
    model_name: str = "model"
) -> None:
    if not gbp_surprise_values or not gbp_decisions:
        print("No GBP decision data to plot.")
        return

    fig, ax = plt.subplots(figsize=(12, 7))
    
    # Prepare data for plotting
    steps = [s for s, _ in gbp_surprise_values]
    surprise_vals = [v for _, v in gbp_surprise_values]
    decisions = [d for _, d in gbp_decisions]

    # Map decisions to colors and markers
    color_map = {'CONSOLIDATE': 'green', 'IGNORE': 'blue', 'REJECT': 'red'}
    marker_map = {'CONSOLIDATE': 'o', 'IGNORE': 'x', 'REJECT': '^'}
    label_map = {'CONSOLIDATE': 'Consolidate (Update)', 'IGNORE': 'Ignore (No Update)', 'REJECT': 'Reject (No Update)'}

    # Plot each decision type
    for decision_type in ['CONSOLIDATE', 'IGNORE', 'REJECT']:
        filtered_steps = [steps[i] for i, d in enumerate(decisions) if d == decision_type]
        filtered_surprise = [surprise_vals[i] for i, d in enumerate(decisions) if d == decision_type]
        
        if filtered_steps:
            ax.scatter(
                filtered_steps,
                filtered_surprise,
                color=color_map[decision_type],
                marker=marker_map[decision_type],
                label=label_map[decision_type],
                alpha=0.6
            )

    ax.set_title(f'GBP Decision vs. Surprise over Steps ({model_name})')
    ax.set_xlabel('Global Steps')
    ax.set_ylabel('Surprise (Gradient Norm)')
    ax.legend()
    ax.grid(True, linestyle='--', alpha=0.7)

    current_date = datetime.now().strftime("%Y%m%d")
    file_name = f"{current_date}-{model_name}-GBP_Decisions.png"
    plt.savefig(os.path.join(output_dir, file_name))
    plt.close()

def plot_metrics(
    step_metrics: Dict[str, List[Tuple[int, float]]],
    epoch_metrics: Dict[str, List[Tuple[int, float]]],
    output_dir: str,
    model_name: str = "model",
    gbp_surprise_values: Optional[List[Tuple[int, float]]] = None,
    gbp_decisions: Optional[List[Tuple[int, str]]] = None
) -> None:
    
    # Determine the number of subplots needed
    num_plots = len(['loss', 'acc', 'pi', 'surprise', 'tau'])
    if gbp_surprise_values and gbp_decisions:
        num_plots += 1 # Add one for the GBP decision plot

    # Adjust figsize based on number of plots
    rows = (num_plots + 1) // 2 # 2 columns
    fig = plt.figure(figsize=(10 * 2, 5 * rows)) # Adjust width and height per plot

    fig.suptitle(model_name, fontsize=16)
    
    metric_types = ['loss', 'acc', 'pi', 'surprise', 'tau']
    plot_titles = {
        'loss': 'Loss',
        'acc': 'Accuracy',
        'pi': 'Predictive Integrity (PI)',
        'surprise': 'Surprise (Gradient Norm)',
        'tau': 'Tau (Entropy)'
    }
    y_labels = {
        'loss': 'Loss',
        'acc': 'Accuracy (%)',
        'pi': 'PI Score',
        'surprise': 'Surprise',
        'tau': 'Tau'
    }

    for i, metric_type in enumerate(metric_types):
        plt.subplot(rows, 2, i + 1)
        
        # Plot step-level data (only for training)
        train_key = f'train_{metric_type}'
        if train_key in step_metrics and step_metrics[train_key]:
            steps, values = zip(*step_metrics[train_key])
            plt.plot(steps, values, alpha=0.5, label=f'Train {metric_type} (Step)')

        # Plot epoch-level data
        for key, data in epoch_metrics.items():
            if metric_type in key and data:
                steps, values = zip(*data)
                label = key.replace('_', ' ').title()
                plt.plot(steps, values, marker='o', linestyle='-', label=f'{label} (Epoch Avg)')

        plt.title(f'{plot_titles[metric_type]} over Steps')
        plt.xlabel('Global Steps')
        plt.ylabel(y_labels[metric_type])
        plt.legend()
        plt.grid(True)

    # Add the GBP decision plot if data is available
    if gbp_surprise_values and gbp_decisions:
        plt.subplot(rows, 2, num_plots) # Place it in the last subplot position
        plot_gbp_decisions_internal(
            plt.gca(), # Get current axes
            gbp_surprise_values,
            gbp_decisions,
            model_name
        )

    plt.tight_layout(rect=[0, 0.03, 1, 0.95]) # Adjust layout to make room for suptitle
    
    current_date = datetime.now().strftime("%Y%m%d")
    file_name = f"{current_date}-{model_name}.png"
    plt.savefig(os.path.join(output_dir, file_name))
    plt.close()

# Internal helper for plotting GBP decisions on a given axes
def plot_gbp_decisions_internal(
    ax,
    gbp_surprise_values: List[Tuple[int, float]],
    gbp_decisions: List[Tuple[int, str]],
    model_name: str
) -> None:
    
    # Prepare data for plotting
    steps = [s for s, _ in gbp_surprise_values]
    surprise_vals = [v for _, v in gbp_surprise_values]
    decisions = [d for _, d in gbp_decisions]

    # Map decisions to colors and markers
    color_map = {'CONSOLIDATE': 'green', 'IGNORE': 'blue', 'REJECT': 'red'}
    marker_map = {'CONSOLIDATE': 'o', 'IGNORE': 'x', 'REJECT': '^'}
    label_map = {'CONSOLIDATE': 'Consolidate (Update)', 'IGNORE': 'Ignore (No Update)', 'REJECT': 'Reject (No Update)'}

    # Plot each decision type
    for decision_type in ['CONSOLIDATE', 'IGNORE', 'REJECT']:
        filtered_steps = [steps[i] for i, d in enumerate(decisions) if d == decision_type]
        filtered_surprise = [surprise_vals[i] for i, d in enumerate(decisions) if d == decision_type]
        
        if filtered_steps:
            ax.scatter(
                filtered_steps,
                filtered_surprise,
                color=color_map[decision_type],
                marker=marker_map[decision_type],
                label=label_map[decision_type],
                alpha=0.6
            )

    ax.set_title(f'GBP Decision vs. Surprise over Steps')
    ax.set_xlabel('Global Steps')
    ax.set_ylabel('Surprise (Gradient Norm)')
    ax.legend()
    ax.grid(True, linestyle='--', alpha=0.7)



import torch
import torch.nn as nn

class PatchEmbedding(nn.Module):
    """
    Splits an image into patches and embeds them.
    
    Args:
        img_size (int): Size of the input image (e.g., 32 for CIFAR-10).
        patch_size (int): Size of each patch (e.g., 4 for CIFAR-10).
        in_channels (int): Number of input channels (e.g., 3 for RGB).
        embed_dim (int): The embedding dimension for each patch.
    """
    def __init__(self, img_size=32, patch_size=4, in_channels=3, embed_dim=768):
        super().__init__()
        self.img_size = img_size
        self.patch_size = patch_size
        self.n_patches = (img_size // patch_size) ** 2
        
        self.proj = nn.Conv2d(
            in_channels,
            embed_dim,
            kernel_size=patch_size,
            stride=patch_size,
        )

    def forward(self, x):
        x = self.proj(x)  # (B, E, P, P)
        x = x.flatten(2) # (B, E, N)
        x = x.transpose(1, 2) # (B, N, E)
        return x

class TransformerBlock(nn.Module):
    """A standard Transformer block."""
    def __init__(self, embed_dim, num_heads, mlp_dim, dropout=0.1):
        super().__init__()
        self.norm1 = nn.LayerNorm(embed_dim)
        self.attn = nn.MultiheadAttention(embed_dim, num_heads, dropout=dropout, batch_first=True)
        self.norm2 = nn.LayerNorm(embed_dim)
        self.mlp = nn.Sequential(
            nn.Linear(embed_dim, mlp_dim),
            nn.GELU(),
            nn.Dropout(dropout),
            nn.Linear(mlp_dim, embed_dim),
            nn.Dropout(dropout)
        )
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        # Attention part
        x_norm = self.norm1(x)
        attn_output, _ = self.attn(x_norm, x_norm, x_norm)
        x = x + self.dropout(attn_output)
        
        # MLP part
        x_norm = self.norm2(x)
        mlp_output = self.mlp(x_norm)
        x = x + self.dropout(mlp_output)
        
        return x

class VisionTransformer(nn.Module):
    """
    A Vision Transformer model customized for smaller images like CIFAR-10.
    """
    def __init__(self, img_size=32, patch_size=4, in_channels=3, num_classes=10,
                 embed_dim=128, depth=6, num_heads=4, mlp_dim=256, dropout=0.1):
        super().__init__()
        
        self.patch_embed = PatchEmbedding(img_size, patch_size, in_channels, embed_dim)
        self.cls_token = nn.Parameter(torch.zeros(1, 1, embed_dim))
        self.pos_embed = nn.Parameter(torch.zeros(1, 1 + self.patch_embed.n_patches, embed_dim))
        self.pos_drop = nn.Dropout(p=dropout)

        self.blocks = nn.ModuleList([
            TransformerBlock(embed_dim, num_heads, mlp_dim, dropout)
            for _ in range(depth)
        ])

        self.norm = nn.LayerNorm(embed_dim)
        self.head = nn.Linear(embed_dim, num_classes)

    def forward(self, x):
        B = x.shape[0]
        x = self.patch_embed(x)
        
        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)
        
        x = x + self.pos_embed
        x = self.pos_drop(x)
        
        for block in self.blocks:
            x = block(x)
            
        x = self.norm(x)
        
        # Get CLS token for classification
        cls_token_final = x[:, 0]
        logits = self.head(cls_token_final)
        
        return logits

class MoELayer(nn.Module):
    """A Mixture of Experts layer."""
    def __init__(self, in_features, hidden_features, out_features, num_experts, top_k=1):
        super().__init__()
        self.num_experts = num_experts
        self.top_k = top_k
        self.gating = nn.Linear(in_features, num_experts)
        
        self.experts = nn.ModuleList([
            nn.Sequential(
                nn.Linear(in_features, hidden_features),
                nn.GELU(),
                nn.Linear(hidden_features, out_features)
            ) for _ in range(num_experts)
        ])

    def forward(self, x):
        batch_size, num_tokens, in_features = x.shape
        x_flat = x.reshape(-1, in_features)

        gating_logits = self.gating(x_flat)
        
        weights, indices = torch.topk(gating_logits, self.top_k, dim=-1)
        weights = torch.softmax(weights, dim=-1).to(x.dtype)
        
        dispatch_tensor = torch.zeros(x_flat.shape[0], self.num_experts, device=x.device, dtype=x.dtype)
        dispatch_tensor.scatter_(1, indices, weights)
        
        # Non-einsum implementation to avoid inplace errors with torch.compile
        dispatched_x = dispatch_tensor.unsqueeze(-1) * x_flat.unsqueeze(1)
        
        expert_outputs_list = []
        for i in range(self.num_experts):
            expert_input_i = dispatched_x[:, i, :]
            expert_output_i = self.experts[i](expert_input_i)
            expert_outputs_list.append(expert_output_i)
            
        expert_outputs = torch.stack(expert_outputs_list, dim=1)

        # Non-einsum implementation for combining outputs
        combined_output = (dispatch_tensor.unsqueeze(-1) * expert_outputs).sum(dim=1)
        
        final_output = combined_output.reshape(batch_size, num_tokens, -1)
        
        # For GBP or other analysis, we might need to know which experts were activated.
        # We return the raw gating logits for flexibility.
        return final_output, gating_logits

class MoEVisionTransformer(VisionTransformer):
    """A Vision Transformer with MoE layers replacing the MLP layers."""
    def __init__(self, num_experts=4, top_k=1, **kwargs):
        # Pop 'model_type' as it's not a parameter for the parent VisionTransformer
        kwargs.pop('model_type', None)
        super().__init__(**kwargs)

        # Extract necessary dimensions from kwargs
        embed_dim = kwargs.get('embed_dim', 128)
        mlp_dim = kwargs.get('mlp_dim', 256)
        depth = kwargs.get('depth', 6)

        # Replace MLP blocks with MoE layers
        for i in range(depth):
            self.blocks[i].mlp = MoELayer(
                in_features=embed_dim,
                hidden_features=mlp_dim,
                out_features=embed_dim,
                num_experts=num_experts,
                top_k=top_k
            )

    def forward(self, x):
        B = x.shape[0]
        x = self.patch_embed(x)
        
        cls_tokens = self.cls_token.expand(B, -1, -1)
        x = torch.cat((cls_tokens, x), dim=1)
        
        x = x + self.pos_embed
        x = self.pos_drop(x)
        
        all_gating_logits = []
        for block in self.blocks:
            # Attention part is the same
            x_norm1 = block.norm1(x)
            attn_output, _ = block.attn(x_norm1, x_norm1, x_norm1)
            x = x + block.dropout(attn_output)
            
            # MoE MLP part
            x_norm2 = block.norm2(x)
            mlp_output, gating_logits = block.mlp(x_norm2)
            x = x + block.dropout(mlp_output)
            all_gating_logits.append(gating_logits)
            
        x = self.norm(x)
        
        cls_token_final = x[:, 0]
        logits = self.head(cls_token_final)
        
        # Return logits and gating decisions for analysis
        return logits, all_gating_logits

    def zero_inactive_expert_grads(self, all_gating_logits):
        """
        Zeros out the gradients for inactive experts.
        This is the core of the FFB's selective update mechanism.
        """
        with torch.no_grad():
            for i, block in enumerate(self.blocks):
                if hasattr(block.mlp, 'experts'):
                    gating_logits_block = all_gating_logits[i]
                    # Assuming gating_logits_block is on the same device as the model
                    device = gating_logits_block.device 
                    
                    _, top_indices = torch.topk(gating_logits_block, block.mlp.top_k, dim=-1)
                    
                    active_experts_mask = torch.zeros(block.mlp.num_experts, dtype=torch.bool, device=device)
                    active_experts_mask[top_indices.unique()] = True
                    
                    for expert_idx, expert_layer in enumerate(block.mlp.experts):
                        if not active_experts_mask[expert_idx]:
                            for param in expert_layer.parameters():
                                if param.grad is not None:
                                    param.grad.zero_()



# GBP MoE ViT Configuration

# Model parameters
model_config = {
    'model_type': 'gbp_moe',
    'img_size': 32,
    'patch_size': 4,
    'in_channels': 3,
    'num_classes': 10,
    'embed_dim': 128,
    'depth': 6,
    'num_heads': 4,
    'mlp_dim': 32,  # mlp_dim for each expert
    'dropout': 0.1,
    'num_experts': 16, # More experts for GBP
    'top_k': 4,       # Select top 2 experts
}

# Training parameters
train_config = {
    'epochs': 20,
    'batch_size': 512,
    'accumulation_steps': 2,
    'learning_rate': 1e-3,
    'weight_decay': 1e-4,
    'output_dir': 'output/ViT/',
    'train_fn_kwargs': {
        'initial_surprise_ema': 3.0  # Bootstrap EMA for GBP
    }
}

# PI Monitor parameters
pi_config = {
    'alpha': 1.0,
    'gamma': 0.5,
}
